/**
 * Created by Ludo Pulles on 2017-12-26.
 */

/*******************************************************************************
 * Flippo is the object which manages the flippo visualisation.
 * This part makes sure that the board transitions can be made.
 * When the board is changed, it sends the board to all its controllers.
 ******************************************************************************/

// w: contains white piece
// b: contains black piece
// W: white placed piece here
// B: black placed piece here
// x: black turned to white
// c: white turned to black
//  : empty

String.prototype.setCharAt = function(index, ch) {
	if (0 <= index && index < this.length)
		return this.substr(0, index) + ch + this.substr(index + 1);
	return this.toString();
}

function isWhite(board, row, col) { return "wWx".indexOf(board.charAt(8 * row + col)) >= 0; }
function isBlack(board, row, col) { return "bBc".indexOf(board.charAt(8 * row + col)) >= 0; }

function processFlips(board, row, col)
{
	var dx = [ 1, 1, 1, 0, -1, -1, -1, 0 ];
	var dy = [ -1, 0, 1, 1, 1, 0, -1, -1 ];

	function isValid(r, c) {
		if (r < 0 || r >= 8) return false;
		if (c < 0 || c >= 8) return false;
		return board.charAt(8 * r + c) != ' ';
	}

	function isCurrentPlayer(r, c) {
		var ch1 = board.charAt(8 * row + col);
		if (ch1 == 'W') {
			return isWhite(board, r, c);
		} else {
			if (ch1 != 'B')
				console.error('Strange character: ' + ch1);
			return isBlack(board, r, c);
		}
	}

	for (var dir = 0; dir < 8; dir++) {
		var stx = row, sty = col;
		while (isValid(stx, sty)) {
			stx += dx[dir];
			sty += dy[dir];
		}
		do {
			stx -= dx[dir];
			sty -= dy[dir];
		} while (!isCurrentPlayer(stx, sty)); // is already valid

		if (stx == row && sty == col) continue;
		while ( stx -= dx[dir],
				sty -= dy[dir],
				stx != row || sty != col) {
			
			if (!isWhite(board, stx, sty) && !isBlack(board, stx, sty))
				console.error("Neither white nor black at position (" + stx + ", " + sty + ")");
			var ch = isWhite(board, stx, sty) ? 'c' : 'x';
			// flip this stone
			board = board.setCharAt(8 * stx + sty, ch);
		}
	}
	return board;
}

function convertMoves(moves) {
	var lastBoard = "                           wb      bw                           ";

	/* var lastBoard = "";
	for (var i = 0; i < 64; i++) {
		var ch = ' ';
		if (i == 27 || i == 36) ch = 'w';
		if (i == 28 || i == 35) ch = 'b';
		lastBoard += ch;
	}
	console.log(lastBoard); */

	var ret = [];
	ret.push(lastBoard);
	for (var idx in moves) {
		// convert the move into a new board:
		var row = moves[idx].charCodeAt(0) - 'A'.charCodeAt(0);
		var col = moves[idx].charCodeAt(1) - '1'.charCodeAt(0);

		var nextBoard = lastBoard.replace(/W|x/g, 'w').replace(/B|c/g, 'b');
		nextBoard = nextBoard.setCharAt(row * 8 + col, idx % 2 == 0 ? 'W' : 'B');
		nextBoard = processFlips(nextBoard, row, col);

		ret.push(lastBoard = nextBoard);
	}
	return ret;
}

function Flippo(playerNames, moves) {
	moves = convertMoves(moves);

	this.playerNames = playerNames;
	this.char1 = '1'.charCodeAt(0);
	this.charA = 'A'.charCodeAt(0);
	this.size = 8;
	var N = this.size * this.size;
	this.isRunning = false;

	this.moves = [];
	for (var idx in moves) {
		var move = moves[idx], pos = -1;

		if (idx == 0) {
			this.moves.push({ board: move, isStart: true });
			continue;
		}

		if (move.indexOf('W') >= 0) {
			pos = move.indexOf('W');
		} else if (move.indexOf('B') >= 0) {
			pos = move.indexOf('B');
		} else {
			console.warn('no current move found at move ' + move);
		}
		var row = Math.floor(pos / 8), col = pos % 8;
		this.moves.push({
			board: move,
			isStart: false,
			row: row,
			col: col,
			move: String.fromCharCode(this.charA + row) + (col + 1)
		});
	}

	this.board = document.getElementById('game-container');
	this.moveTable = document.getElementById('move-table');

	var showLegends = false;

	// create the board
	var table = document.createElement('table');
	var thead = document.createElement('thead');
	var thr = document.createElement('tr');

	if (showLegends) {
		var emptyTH = document.createElement('th');
		emptyTH.className = "row-label col-label";
		emptyTH.id = "row0 col0";
		thr.appendChild(emptyTH);

		for (var col = 0; col < this.size; col++) {
			var th = document.createElement('th');
			th.innerText = (col + 1);
			th.className = "col-label";
			th.id = "col" + (col + 1);
			thr.appendChild(th);
		}
		thead.appendChild(thr);
		table.appendChild(thead);
	}

	this.cells = new Array(this.size);
	var tbody = document.createElement('tbody');
	for (var row = 0; row < this.size; row++) {
		var tr = document.createElement('tr');

		if (showLegends) {
			var rowName = String.fromCharCode(this.charA + row);
			var rowElem = document.createElement('td');
			rowElem.innerText = rowName;
			rowElem.className = "row-label";
			rowElem.id = "row" + (row + 1);
			tr.appendChild(rowElem);
		}

		this.cells[row] = new Array(this.size);
		for (var col = 0; col < this.size; col++) {
			var cur = document.createElement('td');
			var label = document.createElement('div');
			var stone = document.createElement('div');

			label.innerText = String.fromCharCode(this.charA + row) + (col + 1);

			cur.className = "game-cell";
			label.className = "cell-label";
			stone.className = "stone empty";

			cur.appendChild(label);
			cur.appendChild(stone);
			tr.appendChild(cur);

			this.cells[row][col] = stone;
		}
		tbody.appendChild(tr);
	}
	table.appendChild(tbody);

	this.board.appendChild(table);

	this.curMove = -1;

	this.registerEvents();
	this.buildMoveTable();

	this.setMove(0);
}

// returns whether the move has been processed
Flippo.prototype.setMove = function(move) {
	if (move < 0 || move >= this.moves.length) return false;
	if (move == this.curMove) return true;

	var isNext = move == this.curMove + 1;
	this.curMove = move;

	this.updateSelectedRow();
	this.updateGame(isNext);
	return true;
};

Flippo.prototype.setRunning = function(running) {
	if (this.isRunning == running) return;

	this.isRunning = running;
	this.playPause.innerText = running ? 'Pause' : 'Play';

	var self = this, run = function() {
		if (!self.isRunning) return;

		if (self.setMove(self.curMove + 1)) {
			setTimeout(run, 500);
		} else {
			// last move has been reached.
			self.setRunning(false);
		}
	}
	if (this.isRunning) run();
}

Flippo.prototype.prevMove = function() { this.setRunning(false); this.setMove(this.curMove - 1); };
Flippo.prototype.nextMove = function() { this.setRunning(false); this.setMove(this.curMove + 1); };
Flippo.prototype.setShowFlips = function(isChecked) { this.board.className = isChecked ? "show-flips" : ""; }

Flippo.prototype.registerEvents = function() {
	this.playPause = document.getElementById('play-pause');

	var self = this;
	document.getElementById('first').onclick = function() { self.setRunning(false); self.setMove(0); };
	document.getElementById('last').onclick = function() { self.setRunning(false); self.setMove(self.moves.length - 1); };
	document.getElementById('prev').onclick = function() { self.prevMove(); };
	document.getElementById('next').onclick = function() { self.nextMove(); };
	this.playPause.onclick = function() { self.setRunning(!self.isRunning); };

	document.getElementById('show-flips').onchange = function() { self.setShowFlips(this.checked); };
};


Flippo.prototype.buildMoveTable = function() {
	var table = document.createElement('table'), self = this;
	var theader = document.createElement('tr');
	theader.appendChild(document.createElement('th'));
	for (var i = 0; i < 2; i++) {
		var th = document.createElement('th');

		// cut-off on more than 20 characters in the name
		var playerName = i == 0 ? "white" : "black";
		if (this.playerNames) {
			var name = this.playerNames[i];
			if (name.length > 20) name = name.slice(0, 17) + "...";

			playerName = name + " as " + playerName;
			// playerName = "Player " + (i + 1) + " (" + name + ")";
		}

		th.innerText = playerName;
		theader.appendChild(th);
	}

	var thead = document.createElement('thead');
	thead.appendChild(theader);
	table.appendChild(thead);

	var tbody = document.createElement('tbody');
	for (var i = 0; i < this.moves.length; i++) {
		var tr = document.createElement('tr');
		var moveNr = document.createElement('th');
		moveNr.innerText = i + '.';
		if (i % 2) tr.setAttribute('class', 'zebra');
		tr.appendChild(moveNr);

		var mv = this.moves[i];
		var currentMove = mv.isStart ? "Start" : mv.move;

		for (var j = 0; j < 2; j++) {
			var td = document.createElement('td');
			if (i == 0) { // start move
				if (j == 0) td.innerText = currentMove;
			} else if (currentMove == 'Quit' || (i % 2) != j) {
				td.innerText = currentMove;
			}
			tr.appendChild(td);
		}

		tr.onclick = (function(move) {
			return function() {
				self.setMove(move);
			}
		})(i);

		tbody.appendChild(tr);
	}

	this.moveTableBody = tbody;
	this.moveTable.setAttribute('tabindex', '1');
	this.moveTable.onkeydown = function(event) {
		if (event.keyCode == 38 || event.keyCode == 37) {
			this.prevMove();
		} else if (event.keyCode == 40 || event.keyCode == 39) {
			this.nextMove();
		} else {
			return;
		}
		event.preventDefault();
		event.stopImmediatePropagation();
	};

	table.appendChild(tbody);
	this.moveTable.appendChild(table);
};

Flippo.prototype.updateSelectedRow = function() {
	var lastCurMove = document.getElementById('cur-move');
	var tbl = this.moveTable;

	// reset the id.
	if (lastCurMove) lastCurMove.id = '';

	if (this.curMove >= 0) {
		var curRow = this.moveTableBody.children[this.curMove];
		curRow.id = 'cur-move';

		var topCoord = curRow.offsetTop;
		var bottomCoord = curRow.offsetTop + curRow.offsetHeight - tbl.offsetHeight;
		if (tbl.scrollTop < bottomCoord) {
			tbl.scrollTop = bottomCoord;
		} else if (tbl.scrollTop > topCoord) {
			tbl.scrollTop = topCoord;
		}
	} else {
		tbl.scrollTop = 0;
	}
};

Flippo.prototype.updateGame = function(isNext) {
	var content = this.moves[this.curMove].board;
	for (var row = 0; row < 8; row++) {
		for (var col = 0; col < 8; col++) {
			var idx = 8 * row + col;
			var ch = content.charAt(idx);

			var classes = "stone ";
			if ("wWx".indexOf(ch) >= 0)
				classes += "player0";
			else if ("bBc".indexOf(ch) >= 0)
				classes += "player1";
			else
				classes += "empty";

			var pClasses = "game-cell ";
			if (ch == 'W' || ch == 'B') pClasses += "last-move";
			else if ((ch == 'x' || ch == 'c') && (true || isNext)) pClasses += "flipped";

			var cur = this.cells[row][col];
			cur.className = classes;
			var par = cur.parentElement;
			// console.log(par);
			par.className = pClasses;
		}
	}
};

